/**
 * @(#)tstd2tsid.cpp 2003/02/01
 *
 * @author Ice
 * @version 1.00
 *
 * ICE Team free software group
 *
 * This file is part of TSID2.
 * See README for copyright notice.
 *
 * note: some comments are in the style of Javadoc of Java language!
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 */
 
/** 
 * This program converts the stored TSID data into TSID2 data format
 *
 * The TSID directory is passed and each voice that can be converted to
 * TSID2 data will be then erased from TSID.
 */

#include <tsid/osdepend.h>
#include <tsid/tsid.h>
#include "sidt2.h"
#include "tblist.h"
#include "tsid2.h"

#include <sidplay/utils/SidTuneMod.h>

#include <stdio.h>
#include <string.h>
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <sys/stat.h>

#ifdef HAVE_UNIX
#  include <unistd.h>
#  include <sys/fcntl.h>
#  include <dirent.h>
#elif __BORLANDC__
#  include <dirent.h>
#else
#  include "win32-dirent.h"
#  define S_ISDIR(st_mode) (st_mode & S_IFDIR)
#  define S_ISREG(st_mode) (st_mode & S_IFREG)
#endif

#define NLENGTH(dirent) ((int)strlen((dirent)->d_name))
#define strequ(s1, s2) (strcmp((s1),(s2)) == 0)

#ifdef HAVE_UNIX
#  define TSID2DIR ".tsid2"
#else
#  define TSID2DIR "tsid2"
#endif


// TSID directory where there's the tree
char *TDir;

// Home directory of user
char *homeDir;

// home directory of TSID tree
char *homeTDir;

// directory base of HVSC
char *hvscDir;

// base directory for TSID2;
char *basePath;



/**
 * Get the HVSC dir from enviroment
 *
 * @return true if a valid HVSC was found
 */
bool getHVSCDir() {
  // try to found a valid hvsc
  hvscDir=getenv("HVSC_DIR");
  if (hvscDir==NULL) {
    cerr << "No enviroment for HVSC_DIR set up\n\n";
    cerr << "You must set up the HVSC_DIR to the HVSC directory.\n";
    cerr << "Use the hvsctest program to set up it correctly\n";
    return false;
  }

  char *name=new char[strlen(hvscDir)+30];
  if (name==NULL) {
    cerr << "Error allocating memory\n";
    return false;
  }
  sprintf(name, "%s%c%s%c%s",hvscDir, PATH_SEP, "DOCUMENTS", PATH_SEP, "hv_sids.txt");

  // open the file
  ifstream file(name, ios::in);

  if (!file) {
    cerr << "No valid HVSC found: missing file\n";
    delete [] name;
    return false;
  }

  char ch;

  while (file.get(ch)) {
    if (ch=='s') break;
  }

  double ver;

  file.get(ch);
  if (ch=='e') {

    file >> ver;
    cout << "Found HVSC release " << ver << endl;
  } else {
      cerr << "No valid hv_sidt.txt file found\n";
      delete [] name;
      return false;
   }

  delete [] name;
  return true;
}



/**
  * Convert the data from TSID to TSID2 of the given file
  * Fist we check that the PSID file is in HVSC,
  * then we use it's MD5 for doing the conversion
  *
  * @param name1 name of TSID file
  * @param name2 name of TSID2 file
  */
void convert(char *name1, char *name2) {
  char digest[SIDTUNE_MD5_LENGTH+1];
  unsigned int song;
  unsigned long second;
  unsigned long number;

  // remove the ending 't' from '.sidt' extension
  name2[strlen(name2)-1]='\0';

  // open the file
  ifstream file(name2, ios::in);

  if (!file) {
    // the file did not exists: do not convert
    cerr << "[MISSING] " << name1 << endl;
    return;
  }
  file.close();
 
  SidTuneMod *mod=new SidTuneMod(name2);
  if (mod==NULL) {
    cerr << "[ERROR] creating MD5 of file " << name2 << endl;
    return;
  }

  mod->createMD5(digest); //create the MD5 for this tune
  delete mod;

  // open the SIDT file
  SIDT sid1(name1);

  if (!sid1.read()) {       // read the first file
    cerr << "[ERROR] reading: " << name1 << endl;
  }

  SIDT2 sidt2(digest, name2, basePath);

  sid1.initForGet();
  while (sid1.getNext(&song, &second, &number)) {
    sidt2.add(song, second, number, 0);
  }
  
  ~sid1;
  if (remove(name1)!=0) {
    cerr << "[ERROR] removing the file: " << name1 << endl;
  }

  if (!sidt2.read()) {
    cerr << "[ERROR] reading TSID2 of: " << name1 << endl;
  }
  
  if (!sidt2.write()) {
    cerr << "[ERROR] writing TSID2 of " << name1 << endl;
  }
  ~sidt2;

  cout << "[OK] " << name2 << endl;
  //delete sidt2;
}


/**
 * Recurse through the given directory and manage the founded files.
 * This code is taken and modified from gzip sources
 *
 * @param dir the base TSID directory to recurse in
 * @param dir2 the base directory  for TSID2
 */
void recurse(char *dir, char *dir2) {
  dirent *dp;
  struct stat st;
  DIR    *dirp;
  char   nbuf[MAX_PATH_LEN];
  char   nbuf2[MAX_PATH_LEN*2];
  int    len;
  int    len2;

  cout << "Recurse into directory " << dir << endl;

  dirp=opendir(dir);

  if (dirp==NULL) {
    cerr << dir << " unreadable\n";
    return;
  }

  while ((dp=readdir(dirp)) != NULL) {
    if (strequ(dp->d_name,".") || strequ(dp->d_name,"..")) {
      continue;
    }
    len=strlen(dir);
    if (len+NLENGTH(dp)+1 < MAX_PATH_LEN-1) {
      strcpy(nbuf,dir);
      if (len != 0                 // dir = "" means current dir on Amiga
        ) {
        nbuf[len++]='/';
      }
      strcpy(nbuf+len, dp->d_name);

      stat(nbuf, &st);

      len2=strlen(dir2);
      strcpy(nbuf2,dir2);
      if (len2 != 0                 // dir = "" means current dir on Amiga
        ) {
        nbuf2[len2++]='/';
      }
      strcpy(nbuf2+len2, dp->d_name);

      if (S_ISDIR(st.st_mode)) {
        recurse(nbuf, nbuf2);                       // recurse into directory
      }

      if (S_ISREG(st.st_mode)) {
        convert(nbuf, nbuf2);                       // convert the file to tsid2
      }
    } else {
        cerr << dir << "/" << dp->d_name << ": pathname too long\n";
      }
  }
  closedir(dirp);
}


/**
 * Get home directory of user
 * The private variables homeDir is chaged
 * Note: this code is give by Simon White
 *
 * @return true if operation is ok
 */
bool getHomeDir() {
  char *path = (char *) getenv ("HOME");

  if (!path)
    path = (char *) getenv ("windir");
  if (!path) {
    return false;
  }

  homeDir=new char[strlen(path)+1];
  if (homeDir==NULL) {
    return false;
  }
  strcpy(homeDir, path);
  return true;
}

/**
 * Convert platform specific directory seperator to ansi
 *
 * @param filename the filename to made with '/'
 */
void convertDirSeperator(char *filename) {
#ifndef HAVE_UNIX
  char *p = filename;
  while (*p != '\0') {
    if (*p == PATH_SEP) *p = '/';
    p++;
  }
#endif
}


int main(int argc, char** argv) {
  DIR    *dirp;

  {
    // make sure that the base directory exist
    TSID2 tsid2(TBLIST::DINAMIC, 100, 100, NULL);
    ~tsid2;
  }  

#ifndef HAVE_UNIX
  TDir="tsid_hvsc";
#else
  // base directory is not directly visible in unix system
  TDir=".tsid_hvsc";
#endif

  // read the home directory of user
  if (!getHomeDir()) {
    cerr << "Error: not home directory found\n";
    return 1;
  }

  homeTDir=new char[strlen(homeDir)+strlen(TDir)+2];
  if (homeTDir==NULL) {
      cerr << "Error allocating memory\n";
      return 1;
  }
  sprintf(homeTDir, "%s%c%s", homeDir, PATH_SEP, TDir);

  if (!getHVSCDir()) {
    cout << "Cannot continue without a valid HVSC_DIR set up\n";
    return 2;
  }

  convertDirSeperator(homeTDir);

  basePath=new char[strlen(homeDir)+strlen(TSID2DIR)+2];
    if (basePath==NULL) {
      // no memory error
      cerr << "Error allocating memory\n";
      return 1;
    }
  sprintf(basePath, "%s/%s", homeDir, TSID2DIR);
  convertDirSeperator(basePath);

  // verify if base directory tree is present
  dirp=opendir(homeTDir);
  if (dirp==NULL) {
                                                // not present
    cout << "TSID directory not present\n";
    return 1;
  } else {                                      // present, close handle
      closedir(dirp);
    }
    

  recurse(homeTDir, hvscDir);
  
  //now the realtime list must be rebuilt
  cout << "Rebuild of the dinamic list. This can take long time...\n";
  TBLIST best(TBLIST::DINAMIC, basePath, 100, 100, NULL);
  best.rebuildLists();
  ~best;
}

