/**
 * @(#)tupdate2.cpp 2003/06/16
 *
 * @author Ice
 * @version 1.00
 *
 * ICE Team free software group
 *
 * This file is part of TSID2.
 * See README for copyright notice.
 *
 * note: some comments are in the style of Javadoc of Java language!
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 */

/**
 * This is a program that update the TSID2 tree to be in synch with HVSC 
 * or CGSC upgrades.
 * It uses a custom update file (updatexxx.thvs2) generated with a script from
 * updatexx.hvs and corrected by hand.
 */

#include "tsid2.h"
#include "sidt2.h"

#include <errno.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <dirent.h>

#ifdef HAVE_UNIX
  #include <unistd.h>
  #include <sys/fcntl.h>
#else
  #include <fcntl.h>
  #include <io.h>
  #include <dir.h>
#endif

#define strequ(s1, s2) (strcmp((s1),(s2)) == 0)

// show debug string if DEBUG is defined
#ifdef DEBUG_TUPDATE2
  #define TDEBUG2(a) \
    cerr << "[TUPDATE2] " << a;
#else
  #define TDEBUG2(a) ;
#endif

#ifdef HAVE_UNIX
 #define fromUnixPath(a) \
  ;
#else
  /**
   * Convert path separators from Unix to the other operating system
   * This is not need for Unix system
   *
   * @param path the path to change with os specific path separator
   */
  void fromUnixPath(char *path) {
    for (int i=strlen(path)-1; i>0; i--) {
      if (path[i]=='/') path[i]=PATH_SEP;
    }
  }
#endif

// paramenter field
const int P_NOTHING   = 0x000;  // nothing passed
const int P_HELP      = 0x001;  // help syntax
const int P_INIT      = 0x002;  // init tree
const int P_UPDATE    = 0x004;  // update tree
const int P_BACKUP    = 0x008;  // backup tree
const int P_ERROR     = 0x010;  // error

const int E_OK      = 0;  // no error
const int E_MEMORY  = 1;  // no memory error
const int E_UNREAD  = 2;  // unread directory error
const int E_CHECK   = 3;  // error cheching directory
const int E_ALREADY = 4;  // already initialized
const int E_SYSTEM  = 5;  // error in system external call
const int E_WHO     = 6;  // error in find user
const int E_PARAM   = 7;  // error in paramenter passed
const int E_ENVIR   = 8;  // error in enviroment variable
const int E_MISSING = 9;  // missing file
const int E_NOHVSC  = 10; // no hvsc file found
const int E_REMOVE  = 11; // error removing directory
const int E_FROM    = 12; // missing FROM instruction
const int E_SYNTAX  = 13; // syntax error in parsing
const int E_INIT    = 14; // tree not initialized error
const int E_TO      = 15; // missing TO instructions
const int E_VERSION = 16; // invalid version
const int E_VER     = 17; // missing VER instruction

// last error in program
int lastError=E_OK; // no error

// version of the HVSC related with the TSID tree
char *hvscVer=NULL;

// version of the update apply to tsid tree
char tupdateVer;

// home directory
char *homeDir=NULL;

// tsid tree directory
char *tsid2Dir=NULL;

// name of tsid directory
#ifndef HAVE_UNIX
  char *TDir="tsid2";
#else
  // base directory is not directly visible in unix system
  char *TDir=".tsid2";
#endif


// name of documentation directory
char *TDoc="DOCUMENTS";

// name of the thvs update file
char *updateFile="update.thvs2";

TSID2 tsid2;

/**
 * Instruction DELETE file
 *
 * @param f the file in absolute path
 * #return true if operation is ok
 */
bool instDELETE(char *f) {
  TDEBUG2("DELETE " << f << endl);

    // open the file
  ifstream file(f, ios::in);

  if (!file) {
    // nothing to do
    return true;
  }

  file.close();

  if (remove(f)!=0) {
    cerr << "Error removing the file: " << f << endl;
    return false;
  }
  return true;
}

/**
 * Instruction MOVE file1 to file2
 * If file2 esists, file1 is merged. File1 is then deleted.
 *
 * @param f1 the file to move abs path
 * @param f2 the file where to move abs path
 + @return true if operation is ok
 */
bool instMOVE(char *f1, char *f2) {
  bool ok=true;               // actual state of operation  
  TDEBUG2("MOVE " << f1 << " " << f2 << endl);

  SIDT2 sid1(f1, "#/#", tsid2Dir);
  SIDT2 sid2(f2, "#/#", tsid2Dir);

  unsigned int song;
  unsigned long second;
  unsigned long number;
  unsigned long length;

  if (!sid1.read()) {       // read the first file
    cerr << "Error reading one file: " << f1 << endl;
    ok=false;
  }

  sid1.initForGet();
  while (sid1.getNext(&song, &second, &number, &length)) {
    sid2.add(song, second, number, length);    
  }

  // avoid to write if actually this is empty
  if (!sid2.isEmpty()) {
    if (!sid2.read()) {
      cerr << "Error reading one file: " << f2 << endl;
      ok=false;
    }
    if (!sid2.write()) {
      cerr << "Error writing the file: " << f2 << endl;
      ok=false;
    }
  }  

  //remove file1
  return (ok && instDELETE(f1));
}

/**
 * Instruction SHIFT file, step, from song
 *
 * @param f the file in abs path
 * @param step the step to shift
 * @param from the song from where shifting
 * @return true if operation is ok
 */
bool instSHIFT(char *f, int step, int from) {
  bool ok=true;                   // actual operation state
  TDEBUG2("SHIFT " << f << " " << step << " #" << from << endl);

  SIDT2 sid(f, "#/#", tsid2Dir);

  if (!sid.read()) {
    cerr << "Error reading one file\n";
    ok=false;
  }
  
  // avoid to write if actually this is empty  
  if (!sid.isEmpty()) {
    sid.shift(step, from);
    if (!sid.write()) {
      cerr << "Error writing the file\n";
      ok=false;
    }
  }  
  return ok;
}

/**
 * Instruction SWAP tune n with tune m
 *
 * @param f the file to swap
 * @param n tune to swap with m
 * @param m tune to swap with n
 * @return true if operation is ok
 */
bool instSWAP(char *f, int n, int m) {
  bool ok=true;                   // actual operation state
  TDEBUG2("SWAP " << f << " #" << n << " #" << m << endl);

  SIDT2 sid(f, "#/#", tsid2Dir);

  if (!sid.read()) {
    cerr << "Error reading the file\n";
    ok=false;
  }
  
  // avoid to write if actually this is empty
  if (!sid.isEmpty()) {  
    sid.swap(n,m);

    if (!sid.write()) {
      cerr << "Error writing the file\n";
      ok=false;
    }
  }  
  return ok;
}

/**
 * Instruction GOES file1 tune n to file2 tune m
 * File1 is erased if it become empty.
 *
 * @param f1 the first file name
 * @param f2 the second file name
 * @param n number of tune of first file
 * @param m number of tune of second file
 * @return true if operation is ok
 */
bool instGOES(char *f1, char *f2, unsigned int n, unsigned int m) {
  bool ok=true;
  TDEBUG2("GOES " << f1 << " " << f2 << " #" << n << " #" << m << endl);

  SIDT2 sid1(f1, "#/#", tsid2Dir);
  SIDT2 sid2(f2, "#/#", tsid2Dir);

  unsigned int song;
  unsigned long second;
  unsigned long number;
  unsigned long length;

  if (!sid1.read()) {
    cerr << "Error reading the file" << f1 << endl;
    ok=false;
  }

  //search for the tune to move
  sid1.initForGet();
  while (sid1.getNext(&song, &second, &number, &length)) {
    if (song==n) {
      sid1.remove(n);                             // remove the tune n
      sid2.add(m, second, number, length);        // add to tune m
      break;
    }
  }

  if (!sid1.write()) {
    cerr << "Error writing the file " << f1 << endl;
    ok=false;
  }

  // avoid to write if actually this is empty
  if (!sid2.isEmpty()) {
    if (!sid2.read()) {
      cerr << "Error reading the file " << f2 << endl;
      ok=false;
    }

    if (!sid2.write()) {
      cerr << "Error writing the file " << f2 << endl;
      ok=false;
    }
  }  
  return ok;
}

/**
 * Get parameters summary as an integer value
 *
 * @param argc the numbr of passed string
 * @param argv the strings of parameters
 * @return an integer that summary all the options available
 */
int getParameter(int argc, char** argv) {
  int value=P_NOTHING;                       // returned value;

  // parse command line arguments
  int a=1;
  while ((a<argc) && (argv[a] != NULL))	{
    if (argv[a][0]=='-') {
      int b=1;
      while (argv[a][b]!='\0') {
        switch (argv[a][b]) {
          case 'u':                          // update tree
            value|=P_UPDATE;
            break;
          case 'b':                          // backup tree
            value|=P_BACKUP;
            break;
          case 'h':                          // help
            value|=P_HELP;
            break;
          case 'i':                          // init tree
            value|=P_INIT;
            break;
          default:
            return P_ERROR;
        }
        b++;                                 // next char
      }
    } else {
        // option must be specify with -
        return P_ERROR;
      }

    a++;                                     // next parameter
  }
  return value;
}

/**
 * Backup the tsid tree with a tar.gz or zip file
 *
 * @return true if operation is ok
 */
bool backup() {
  char *tar="tar -czf tsid_backup.tar.gz ";
  char *tarCom=NULL;

  TDEBUG2("Back up of tsid tree directory\n");

  tarCom=new char [strlen(tar)+strlen(tsid2Dir)+1];

  if (tarCom==NULL) {
    cerr << "Error allocating memory\n";
    lastError=E_MEMORY;
    return false;
  }

  sprintf(tarCom, "%s%s", tar, tsid2Dir);
  TDEBUG2("Tar command to execute: " << tarCom << endl);

  // backup with tar/gzip if present
  if (system(tarCom)==0) {
    delete [] tarCom;
    return true;                         // backup is done
  };

  delete [] tarCom;

  //try with zip
  ///not yet implemented
  lastError=E_SYSTEM;
  return false;

  return true;
}

/**
 * Update the TSID tree.
 */
bool update() {
  const int LEN=10;
  const int LENF=256;

  char *name;
  char instr[LEN]   = "";   // instruction name
  char file1[LENF]  = "";   // file name 1
  char file2[LENF]  = "";   // file name 2

  double version    = 0;    // version of the required library

#define FILE_GET(AB) if (file.get()!=AB) { \
                      cerr << "Error parsing update file\n"; \
                      lastError=E_SYNTAX; \
                      return false; \
                    }

  TDEBUG2("Update the tsid2 tree\n");

  //make a backup copy of your tsid2 tree
  if (!backup()) {
    cerr << "I'm not able to create a backup copy of your TSID2 tree\n";
    return false;
  }

  //open the update file from home directory
  name=new char[strlen(homeDir)+strlen(updateFile)+2];
  if (name==NULL) {
    cerr <<"Error allocating memory\n";
    lastError=E_MEMORY;
    return false;
  }
  sprintf(name, "%s%c%s", homeDir, '/', updateFile);

  // open the file
  ifstream file(name, ios::in);

  if (!file) {
    cerr << "The update.thvs2 file is not present in your home directory\n";
    lastError=E_MISSING;
    delete [] name;
    return false;
  }

  // search for the expected VERSION x.x instruction
  file.get(instr, LEN, ' ');
  if (!(strequ(instr,"VERSION"))) {
    cerr << "Missing expected VERSION instruction from tvhs2 file\n";
    lastError=E_VER;
    delete [] name;
    return false;
  }

  FILE_GET(' ');
  file >> version;
  TDEBUG2("VERSION " << version << endl);
  FILE_GET('\n');

  // verify that the version is good
  if (version>(double)tsid2_version) {
    cerr << "You must use a tupdate2 tool of library " << version <<
            " for using this file\n";
    lastError=E_VERSION;
    delete [] name;
    return false;
  }

  char ch;
  int num;
  int song;
  int song2;

  while (file.get(ch)) {
    if (ch!='\n') {
      cerr << "Line not ended as expected\n";
      lastError=E_SYNTAX;
      delete [] name;
      return false;
    }

    file.get(instr, LEN, ' ');

    // is a MOVE instruction?
    if (strequ(instr, "MOVE")) {
      FILE_GET(' ');
      file.get(file1, LENF, ' ');
      FILE_GET(' ');
      file.get(file2, LENF, '\n');

      if (instMOVE(file1, file2)) {
        cout <<   "OK: MOVE " << file1 << endl;
      } else {
          cerr << "ERROR: MOVE " << file1 << endl;
        }
    }

    // is a delete instruction?
    if (strequ(instr, "DELETE")) {
      FILE_GET(' ');
      file.get(file1, LENF, '\n');

      if (instDELETE(file1)) {
        cout << "OK: DELETE " << file1 << endl;
      } else {
          cerr << "ERROR: DELETE " << file1 << endl;
        }
    }

    // is a shift instruction?
    if (strequ(instr, "SHIFT")) {
      FILE_GET(' ');
      file.get(file1, LENF, ' ');
      FILE_GET(' ');

      file >> num;
      FILE_GET(' ');
      FILE_GET('#');
      file >>song;
      
      if (instSHIFT(file1, num, song)) {
        cout << "OK: SHIFT " << file1 << endl;
      } else {
          cerr << "ERROR: SHIFT " << file1 << endl;
        }
    }

    // is a GOES instruction?
    if (strequ(instr, "GOES")) {
      FILE_GET(' ');
      file.get(file1, LENF, ' ');
      FILE_GET(' ');
      file.get(file2, LENF, ' ');
      FILE_GET(' ');
      FILE_GET('#');

      file >> song;

      FILE_GET(' ');
      FILE_GET('#');

      file >> song2;

      if (instGOES(file1, file2, song, song2)) {
        cout << "OK: GOES " << file1 << endl;
      } else {
          cerr << "ERROR: GOES " << file1 << endl;
        }
    }

    // is a SWAP instruction?
    if (strequ(instr, "SWAP")) {
      FILE_GET(' ');
      file.get(file1, LENF, ' ');
      FILE_GET(' ');
      FILE_GET('#');

      file >> song;
      FILE_GET(' ');
      FILE_GET('#');
      file >>song2;

      if (instSWAP(file1, song, song2)) {
        cout << "OK: SWAP " << file1 << endl;
      } else {
          cerr << "ERROR: SWAP " << file1 << endl;
        }
    }

    // is a comment?
    if (strequ(instr, "%")) {
      file.get(file1, LENF, '\n');
      TDEBUG2("Comment: " << file1 << endl);
    }
  }

  delete [] name;
  return true;
}

/**
 * Show a help screen
 */
void showHelp() {
  cout << "Syntax: tupdate2 -<option>\n";
  cout << "Options:\n";
  cout << " -h          display this screen\n";
  cout << " -u          update tsid2 tree\n";
  cout << " -b          backup tsid2 tree\n";
  cout << endl;
}

/**
 * Get some directory of user
 * Global variables are changed
 * Note: this code is based onto Simon White source
 *
 * @return true if operation is ok
 */
bool getDir() {
  // get home directory of user

  char *path = (char *) getenv ("HOME");
  lastError=E_OK;

  if (!path)
      path = (char *) getenv ("windir");
  if (!path) {
      cerr << "Unable to establish home path\n";
      lastError=E_WHO;  // no user information
      return false;
  }

  homeDir=new char[strlen(path)+1];
  if (homeDir==NULL) {
      cerr << "Error allocating memory\n";
      lastError=E_MEMORY;  // no memory available
      return false;
  }
  strcpy(homeDir, path);

  TDEBUG2("Home directory is: " << homeDir << endl);

  // get tsid directory
  tsid2Dir=new char [strlen(homeDir)+strlen(TDir)+2];

  if (tsid2Dir==NULL) {
    cerr << "Error allocating memory\n";
    lastError=E_MEMORY;
    return false;
  }
  sprintf(tsid2Dir, "%s%c%s", homeDir, '/', TDir);

  TDEBUG2("Path to tsid tree is: " << tsid2Dir << endl);

  return true;
}

int main(int argc, char** argv) {
  int result;                                // parameters results

  cout << "TUPDATE2 1.00\n";
  cout << "This program update your TSID2 tree\n";

  // get parameters value
  result=getParameter(argc, argv);

  if ((result & P_ERROR) || (result==P_NOTHING)) {    // there's a error?
    cout << "Invalid or missing parameter option. Try 'tupdate2 -h' for a syntax help\n";
    return E_PARAM;
  }

  if (result & P_HELP) {                     // help option?
    showHelp();
    return E_OK;
  }

  // get some user directories
  if (!getDir()) {
    return lastError;
  }

  if (result & P_BACKUP) {                   // make backup
    if (backup()) return E_OK;
    else return lastError;
  }

  if (result & P_UPDATE) {                   // make update
    if (update()) {
      cout << "Your TSID2 tree in now updated\n";
      
      cout << "Rebuild the best lists. This can take long time...\n";
      TSID2 *t=new TSID2();
      TBLIST *l=t->getTblist();
      l->rebuildLists();      
      delete t;
      cout << "All is now completed.\n";
      
      return E_OK;
    } else return lastError;
  }
}

// _TUpdate.CPP
