/**
 * @(#)tblist.cpp 2002/02/25
 *
 * @author Ice
 * @version 1.02
 *
 * ICE Team free software group
 *
 * This file is part of TSID2.
 * See README for copyright notice.
 *
 * note: some comments are in the style of Javadoc of Java language!
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 */

#ifndef _TBLIST
#define _TBLIST

#include "tblist.h"
#include "sidt2.h"
#include "tsidexaminer2.h"
#include "tsid2.h"

#include <string.h>
#include <stdio.h>
#include <iostream.h>
#include <fstream.h>
#include <limits.h>

#ifdef HAVE_MSWINDOWS
# define snprintf _snprintf
#endif

// the name of the INI file
#define FILE_NAME "best.dat"

const char *TBLIST::ERR_NO_TBLIST_ERROR  = "TBLIST: No error";
const char *TBLIST::ERR_NO_MEMORY        = "TBLIST: Error allocating memory";
const char *TBLIST::ERR_INI_OPEN_ERROR   = "TBLIST: error opening the INI database";
const char *TBLIST::ERR_INI_DELIMS_ERROR = "TBLIST: error setting delims in the database";
const char *TBLIST::ERR_INI_READ_ERROR   = "TBLIST: error reading INI file";
const char *TBLIST::ERR_INI_WRITE_ERROR  = "TBLIST: error writing INI file";
const char *TBLIST::ERR_NO_EXAM_ERROR    = "TBLIST: error creating Examiner";
const char *TBLIST::ERR_ANALISE_ERROR    = "TBLIST: error analising with Examiner";
const char *TBLIST::ERR_NO_SIDT_ERROR    = "TBLIST: error creating Sidt";
const char *TBLIST::ERR_READ_SIDT_ERROR  = "TBLIST: error reading Sidt";

const char *TBLIST::ERROR_LIST[]={
  ERR_NO_SIDT_ERROR,
  ERR_NO_MEMORY,
  ERR_INI_OPEN_ERROR,
  ERR_INI_DELIMS_ERROR,
  ERR_INI_READ_ERROR,
  ERR_INI_WRITE_ERROR,
  ERR_NO_EXAM_ERROR,
  ERR_ANALISE_ERROR,
  ERR_NO_SIDT_ERROR,
  ERR_READ_SIDT_ERROR
};

/**
 * Construct the TBLIST using the passed engine type
 * The path must be with ansi path separator
 *
 * @param engine the type of the engine
 * @param absBasePath the base path where there are the sidt2 files
 * @param number_t the number of time item to store in the list
 * @param number_n the number of number item  to store in the list
 * @param tsid2 a pointer to tsid2 that is using this class
 */
TBLIST::TBLIST(engineType engine, const char *absBasePath,
               unsigned int number_t, unsigned int number_n, TSID2 *tsid2) {
  BDEBUG("Construct the TBLIST: engine="<<engine<<" path="<<absBasePath<<
          " time-size="<<number_t<<" number-size="<<number_n<<"\n");

  versionString="tblist 1.02 by Ice Team Free Software Group (C) 2000-2005";

  // store the passed value
  TBLIST::engine=engine;
  TBLIST::number_t=number_t;
  TBLIST::number_n=number_n;
  TBLIST::tsid2=tsid2;

  lastAbsPath=NULL;
  lastSidName=NULL;

  database=NULL;
  listTime=NULL;
  listNumber=NULL;
  totalTime=0;
  totalNumber=0;

  TBLIST::absBasePath=new char[strlen(absBasePath)+1];
  absFileName=new char[strlen(absBasePath)+strlen(FILE_NAME)+2];

  if ((TBLIST::absBasePath == NULL) || (absFileName == NULL)) {
    lastError=NO_MEMORY;            // no more memory
    tblistStatus=false;             // tblist cannot go away
    return;
  }
  strcpy(TBLIST::absBasePath, absBasePath);
  sprintf(absFileName, "%s/%s", absBasePath, FILE_NAME);

  BDEBUG("absBasePath=" << absBasePath << endl);
  BDEBUG("absFileName=" << absFileName << endl);

  if (!readLists()) {
    //error
  }
}

/**
 * Destruct the TBLIST
 */
TBLIST::~TBLIST() {
  BDEBUG("Desctruct the TBLIST\n");
  if (!writeLists()) {
    //errore
  }
  if (database != NULL) ini_close(database);

  deleteList(listNumber);
  deleteList(listTime);

  if (absBasePath != NULL) delete[] absBasePath;
  if (absFileName != NULL) delete[] absFileName;
  if (lastAbsPath != NULL) delete[] lastAbsPath;
  if (lastSidName != NULL) delete[] lastSidName;
}

/**
 * Set the size of the lists
 *
 * @param sizeTime the size that the list with time must take
 * @param sizeNumber the size that the list with number must take
 * @return true if the operation is ok
 */
bool TBLIST::setSize(unsigned int sizeTime, unsigned int sizeNumber) {
  TSIDExaminer2 *exam=NULL;
  BDEBUG("Set size time to " << sizeTime << "(" << number_t << ") and number to "
                           << sizeNumber << "(" << number_n << ")\n");

  // set the size of the time list
  for (;;) {
    // avoid the 0 or equal length
    if ((sizeTime == 0) || (sizeTime == number_t)) break;

    if (sizeTime < number_t) {
      // the list must be make shorter
      listTime=reduceList(listTime, sizeTime);
      number_t=sizeTime;
      break;
    } else {
        // force tsid2 to flush open sidt2 files
        if (tsid2 != NULL) tsid2->flush();

        // the list must be rebuilded using examiner
        exam=new TSIDExaminer2(absBasePath);
        if (!exam) {
          lastError=NO_EXAM_ERROR;
          tblistStatus=false;
          return false;
        }
        if (!exam->analise()) {
          lastError=ANALISE_ERROR;
          delete exam;
          return false;
        }

        rebuildListT(sizeTime, exam);
        break;
      }
  }

  // set the size of the number list

  // avoid the 0 or equal length
  if ((sizeNumber == 0) || (sizeNumber == number_n)) return true;

  if (sizeNumber < number_n) {
    // the list must be make shorter
    listNumber=reduceList(listNumber, sizeNumber);
    number_n=sizeNumber;
  } else {
      // force tsid2 to flush open sidt2 files
      if (tsid2 != NULL) tsid2->flush();
      if (exam==NULL) {
        // the list must be rebuilded using examiner
        exam=new TSIDExaminer2(absBasePath);
        if (!exam) {
          lastError=NO_EXAM_ERROR;
          tblistStatus=false;
          return false;
        }
        if (!exam->analise()) {
          lastError=ANALISE_ERROR;
          delete exam;
          return false;
        }
      }
      rebuildListN(sizeNumber, exam);
    }
  delete exam;
  return true;
}

/**
 * Rebuild the time list
 *
 * @param sizeTime the size that the list with time must take
 * @param exam an already initialized TSIDExaminer2 object
 */
void TBLIST::rebuildListT(unsigned int sizeTime, TSIDExaminer2 *exam) {
  BDEBUG("Rebuild list time with size: " << sizeTime << endl);
  TSIDExaminer2::TRec rec;

  exam->sort(TSIDExaminer2::TIME);

  deleteList(listTime);
  listTime=createEmptyList(sizeTime);
  exam->initForGet();

  for (unsigned int i=0; i<sizeTime; i++) {
    if (exam->getNext(&rec)) {
      addList(listTime, rec.MD5, rec.tune, 0, true); // 0 because addList read the value again
      if (rec.path != NULL) delete[] rec.path;
      if (rec.name != NULL) delete[] rec.name;
    } else {
        break;
      }
  }
  number_t=sizeTime;
}

/**
 * Rebuild the number list
 *
 * @param sizeNumber the size that the list with number must take
 * @param exam an already initialized TSIDExaminer2 object
 */
void TBLIST::rebuildListN(unsigned int sizeNumber, TSIDExaminer2 *exam) {
  BDEBUG("Rebuild list number with size: " << sizeNumber << endl);
  TSIDExaminer2::TRec rec;

  exam->sort(TSIDExaminer2::NUMBER);

  deleteList(listNumber);
  listNumber=createEmptyList(sizeNumber);
  exam->initForGet();

  for (unsigned int i=0; i<sizeNumber; i++) {
    if (exam->getNext(&rec)) {
      addList(listNumber, rec.MD5, rec.tune, 0, false); // 0 because addList read the value again
      if (rec.path != NULL) delete[] rec.path;
      if (rec.name != NULL) delete[] rec.name;
     } else {
         break;
       }
  }
  number_n=sizeNumber;
}

/**
 * Rebuild the two lists by scanning all the stored files
 *
 * @return true if the operation is ok
 */
bool TBLIST::rebuildLists() {
  TSIDExaminer2 *exam=NULL;

  BDEBUG("Rebuild the two lists\n");

  // force tsid2 to flush open sidt2 files
  if (tsid2 != NULL) tsid2->flush();

  // the list must be rebuilded using examiner
  exam=new TSIDExaminer2(absBasePath);
  if (!exam) {
    BDEBUG("Error creating examiner\n");
    lastError=NO_EXAM_ERROR;
    tblistStatus=false;      // cannot continue
    return false;
  }
  if (!exam->analise()) {
    BDEBUG("Error analising with examiner\n");
    lastError=ANALISE_ERROR;
    delete exam;
    return false;
  }

  rebuildListT(number_t, exam);
  rebuildListN(number_n, exam);
  totalNumber=exam->getTotalNumber();
  totalTime=exam->getTotalTime();
  delete exam;
  return true;
}

/**
 * Create an empty list with the passed size
 *
 * @param size the size of the list
 * @return the pointer to the list
 */
TBLIST::tblistList* TBLIST::createEmptyList(unsigned int size) {
  BDEBUG("Create an empty list of " << size << " elements\n");
  tblistList *tmp;

  //assign memory
  tmp=(tblistList *)new char[2*sizeof(unsigned int)+sizeof(tblistRecord*)*size];
  if (tmp==NULL) {
    lastError=NO_MEMORY;            // no more memory
    tblistStatus=false;             // tblist cannot go away
    return NULL;
  }

  // fill the elemens to point to null values
  tmp->number=0;
  tmp->size=size;
  for (unsigned int i=0; i<size; i++) {
    tmp->item[i]=NULL;
  }

  return tmp;
}

/**
 * Delete the passed list by free the used memory
 *
 * @param list the pointer to the list
 */
void TBLIST::deleteList(tblistList *list) {
  BDEBUG("Delete the passed list...\n");
  if (list == NULL) return;

  BDEBUG("..that has " << list->size << " elements\n");

  // scan all the list, not limited to list->number
  for (unsigned int i=0; i<list->size; i++) {
    if (list->item[i] != NULL) {
      if (list->item[i]->MD5 != NULL) delete[] list->item[i]->MD5;
      if (list->item[i]->path != NULL) delete[] list->item[i]->path;
      if (list->item[i]->name != NULL) delete[] list->item[i]->name;
      delete list->item[i];
    }
  }
  delete[] list;
}


/**
 * Read the lists info from the file
 *
 * @return true if the operation is ok
 */
bool TBLIST::readLists() {
  BDEBUG("Read the best lists from the database\n");
  int tmp;
  bool status;  // actual status

  // create two empty list
  listTime=createEmptyList(number_t);
  listNumber=createEmptyList(number_n);
  if (!this) return false;

  // test if the files exist
  { // open the file
    ifstream file(absFileName, ios::in);

    if (!file) {
      // nothing to do
      BDEBUG("Nothing to do: file not exist\n");
      return true;
    }
  }

  database=ini_open(absFileName,"w",";");
  if (!database) {
    BDEBUG("Error opening the INI database for the " << absFileName);
    lastError=INI_OPEN_ERROR;
    return false;
  }
  if (ini_listDelims(database, " ") < 0) {
    BDEBUG("Error setting delimiters in the INI database");
    lastError=INI_DELIMS_ERROR;
    return false;
  }

  // read the total time/number of listen
  ini_locateHeading(database, "TOTAL");
  ini_locateKey(database, "time");
  // if this key is present: add to the vector
  if (!ini_readInt(database, &tmp))
      totalTime=(unsigned int)tmp;
  BDEBUG("Total time: " << totalTime << endl);

  ini_locateKey(database, "number");
  // if this key is present: add to the vector
  if (!ini_readInt(database, &tmp))
      totalNumber=(unsigned int)tmp;
  BDEBUG("Total number: " << totalNumber << endl);

  status=readList("HOTLIST_T", listTime, number_t);
  status=status && readList("HOTLIST_N", listNumber, number_n);
  return status;
}

/**
 * Read the item list from the file
 *
 * @param heading the heading string
 * @param list the list to complete with the entries
 * @param maxSize the max size allocated for this object
 * @return true if the operation is ok
 */
bool TBLIST::readList(char *heading, tblistList *list, unsigned int maxSize) {
  BDEBUG("Read the list " << heading << endl);
  unsigned int size;               // size readed from the file
  int tmp;
  char key[20]="";                 // for converting int to string
  int len;                         // length of the string to read
  tblistRecord *rec;
  
  lastError=NO_TBLIST_ERROR;

  // read the list of best listen time
  ini_locateHeading(database, heading);
  ini_locateKey(database, "tunes");

  // this key is present: read it
  if (!ini_readInt(database, &tmp))
      size=(unsigned int)tmp;
  else {
      // this info is not present: assume the list empty
      return true;
  }

  if (size > maxSize) {
    BDEBUG("Found " << size << " instead of " << maxSize << endl);
    // the values should be equals, but if it is big, use the max value
    size = maxSize;
  }

  for (unsigned int i=1; i<=size; i++) {
    //create the new memory for the record
    rec=new tblistRecord;
    if (rec == NULL) {
      BDEBUG("Error: no memory\n");
      lastError=NO_MEMORY;              // no more memory
      tblistStatus=false;               // tblist cannot go away
      return false;
    }

    snprintf(key, 20, "%d", i);
    if (ini_locateKey(database, key) == 0) {
      len = ini_dataLength(database)+1;
      // this MD5 is present: read it
      rec->MD5 = new char[len];
      if (rec->MD5 == NULL) {
        BDEBUG("Error: no memory\n");
        lastError=NO_MEMORY;              // no more memory
        tblistStatus=false;               // tblist cannot go away
        return false;
      }
      if (ini_readString(database, rec->MD5, len)<0) {
        BDEBUG("Error reading string of length: " << len << endl);
        lastError=INI_READ_ERROR;
        delete rec;
        continue;                         // try to go ahead
      }

      // now read the subtune
      if (ini_readInt(database, &tmp)<0) {
        BDEBUG("Error reading int\n");
        lastError=INI_READ_ERROR;
        delete rec;
        continue;                         // try to go ahead
      }
      rec->tune=(unsigned int)tmp;

      // fill the other fields with the expected value according with
      // the engine to use
      switch (engine) {
        case STATIC:
          //rec->cached=true;
          readRecord(rec);
          break;
        case DINAMIC:
          rec->cached=false;    // the rest of the record is not cached
          rec->path=NULL;
          rec->name=NULL;
          break;
      }
      list->item[list->number++]=rec;
    } else delete rec;  // delete this not used record
  }
  BDEBUG("read " << list->number << " elements\n");
  return lastError==NO_TBLIST_ERROR;
}

/**
 * Read the record of the list from the file
 * The MD5 and tune field of the record must be filled with the data to read
 * before calling this routine
 *
 * @param rec the record to read in
 * @return true if the operation is ok
 */
bool TBLIST::readRecord(tblistRecord *rec) {
  // we read the info using SIDT2
  // a good solution was to store all the used SIDT2 in a pull for caching
  // the value, so if we search for the same MD5 but a different tune
  // the value is already loaded.
  // at the moment I leave this simple solution and will see how it works
  BDEBUG("Read the record with SIDT2\n");

  sidtRecord srec;
  SIDT2 *sidt=new SIDT2(rec->MD5, "#/#", absBasePath);

  // check sidt status
  if (!sidt) {
    // error, cannot continue
    BDEBUG("Error creating SIDT2\n")
    lastError=NO_SIDT_ERROR;
    tblistStatus=false;
    return false;
  }

  // read the information
  if (!sidt->read()) {
    // error reading the sidt file
    BDEBUG("Error reading from SIDT2\n");
    lastError=READ_SIDT_ERROR;
    return false;
  }

  sidt->getTune(rec->tune, &srec);

  rec->second=srec.second;
  rec->number=srec.number;

  // insert readed path/name of the sid file
  rec->path=sidt->getReadSidPath();
  rec->name=sidt->getReadSidName();
  
  // now the record is cached
  rec->cached=true;

  delete sidt;
  return true;
}

/**
 * Reduce the size of the passed list by creating a new short list.
 * The passed list will be erased
 *
 * @param list the list to reduce
 * @param size the new shorter size
 * @return a new reduced list
 */
TBLIST::tblistList* TBLIST::reduceList(tblistList *list, unsigned int size) {
  BDEBUG("Reduce list to size=" << size << endl);
  unsigned int i;
  tblistList *tmp=createEmptyList(size);

  tmp->number=0;
  tmp->size=size;

  // copy the element in the new list
  for (i=0; i<size; i++) {
    if (list->item[i] != NULL) {
      tmp->item[i] = list->item[i];
      tmp->number++;
    }
  }

  // delete the element of the old list that are not copied
  for (i=size; i<list->number; i++) {
    if (list->item[i] != NULL) {
      if (list->item[i]->MD5 != NULL) delete[] list->item[i]->MD5;
      if (list->item[i]->path != NULL) delete[] list->item[i]->path;
      if (list->item[i]->name != NULL) delete[] list->item[i]->name;
      delete list->item[i];
    }
  }

  // free the memory of the copied list
  delete list;

  
  return tmp;
}

/**
 * Write the lists info to the file
 *
 * @return true if the operation is ok
 */
bool TBLIST::writeLists() {
  BDEBUG("Write the lists to the INI file\n");

  // is the database not open -> probably was empty
  if (database == NULL) {
    //open/create it
    database=ini_open(absFileName,"w",";");
    if (!database) {
      BDEBUG("Error opening the INI database for the " << absFileName);
      lastError=INI_OPEN_ERROR;
      return false;
    }
    if (ini_listDelims(database, " ") < 0) {
      BDEBUG("Error setting delimiters in the INI database");
      lastError=INI_DELIMS_ERROR;
      return false;
    }
  }

  // insert total statistic value

  // delete old value
  ini_locateHeading(database, "TOTAL");
  ini_deleteHeading(database);

  // insert new one
  ini_locateHeading(database, "TOTAL");
  ini_locateKey(database, "time");

  BDEBUG("Writing total time= " << totalTime << endl);
  if (ini_writeInt(database, totalTime)<0) {
    BDEBUG("Error writing the int: " << totalTime << endl);
    lastError=INI_WRITE_ERROR;
    return false;
  }

  ini_locateKey(database, "number");

  BDEBUG("Writing total number= " << totalNumber << endl);  
  if (ini_writeInt(database, totalNumber)<0) {
    BDEBUG("Error writing the int: " << totalNumber << endl);
    lastError=INI_WRITE_ERROR;
    return false;
  }

  if (!writeList(listTime, "HOTLIST_T")) return false;
  if (!writeList(listNumber, "HOTLIST_N")) return false;
  return true;
}

/**
 * Write the value of the passed list in the database file
 *
 * @param list the pointer to the list
 * @param head the heading to write in the database
 * @return true if the operation is ok
 */
bool TBLIST::writeList(tblistList *list, char *head) {
#define DIMT 40
  BDEBUG("Write the " << head << " heading of the list\n");
  char tmp[DIMT];

  // delete old heading value
  ini_locateHeading(database, head);
  ini_deleteHeading(database);

  // insert new one
  ini_locateHeading(database, head);

  // insert total number of entries
  ini_locateKey(database, "tunes");

  if (ini_writeInt(database, list->number)<0) {
    BDEBUG("Error writing the int: " << list->number << endl);
    lastError=INI_WRITE_ERROR;
    return false;
  }

  // scan all the vector
  for (unsigned int i=0; i<list->size; i++) {
    if (list->item[i] == NULL) continue;      // if not present, do nothing
    if (list->item[i]->MD5 == NULL) {
      BDEBUG("Corrupt data???\n");
      continue; // corrupt data?, do nothing
    }
    snprintf(tmp, (size_t)DIMT, "%u", i+1);
    ini_locateKey(database, tmp);
    snprintf(tmp, (size_t)DIMT, "%s %u", list->item[i]->MD5, list->item[i]->tune);

    if (ini_writeString(database, tmp)<0) {
      BDEBUG("Error writing the string: " << tmp << endl);
      lastError=INI_WRITE_ERROR;
      return false;
    }
  }

  return true;
}


/**
 * Add the passed listen time to the top lists
 *
 * @param MD5 the MD5 of the file tune
 * @param tune the subtune
 * @param second the listen second to add
 * @param number the number of listen time to add
 */
void TBLIST::add(const char *MD5, unsigned int tune, unsigned long second,
                 unsigned long number) {
  BDEBUG("Add " << MD5 << " tune=" << tune << " second=" << second <<
         " number=" << number << endl);

  // update total statistic
  totalTime+=second;
  totalNumber+=number;
  addList(listTime, MD5, tune, second, true);
  BDEBUG("Now there are " << listTime->number << " elements\n");
  addList(listNumber, MD5, tune, number, false);
  BDEBUG("Now there are " << listNumber->number << " elements\n");
}

/**
 * Add the passed listen time to the top lists
 *
 * @param MD5 the MD5 of the file tune
 * @param rec a record of information
 */
void TBLIST::add(const char *MD5, sidtRecord rec) {
  add(MD5, rec.tune, rec.second, rec.number);
}

/**
 * Add the passed rate to the given list at the right position
 *
 * @param list the list to update
 * @param MD5 the MD5 of the file
 * @param tune the subtune to update
 * @param rate the rate to add to the current value
 * @param isTime true if this is the time to update
 * @return true if the operation is ok
 */
bool TBLIST::addList(tblistList *list, const char *MD5, unsigned int tune,
                     unsigned long int rate, bool isTime) {
  BDEBUG("Add " << rate << " for  #" << tune << "-" << MD5 <<" to one list with "
         << list->number <<" elements\n");
  tblistRecord *rec=NULL;

  //security action
  if (list==NULL) {
    // error, cannot continue
    return true;
  }

  //find if the MD5 / tune is already in the list
  // we go from the top of the list
  for (unsigned int i=0; i<list->number; i++) {
    if (list->item[i] != NULL) {  // be sure that exist (just for security)
      if ( (strcmp(list->item[i]->MD5, MD5) ==0) && (list->item[i]->tune == tune)) {
        // tune is already in list
        rec=list->item[i];               // remember the pointer
        if (rec->cached==false) readRecord(rec);
        list->item[i]=NULL;              // erase the element, because it may go upper
        if (isTime) rec->second+=rate;   // update his rate
        else rec->number+=rate;

        updateList(list, rec, i, isTime, true);
        return true;
      }
    } else {
        BDEBUG("This may be not called!\n");
      }
  }
  // tune was not in the list; read his actual information
  // we suppose that the sidt2 file is not already updated with
  // the passed information.
  // The class tsid2 does this, so go away

  SIDT2 *sidt=new SIDT2(MD5, "#/#", absBasePath);
  if (!sidt) {
    // error initializing
    lastError=NO_SIDT_ERROR;
    tblistStatus=false;
    return false;
  }

  // read the information
  if (!sidt->read()) {
    // error reading
    lastError=READ_SIDT_ERROR;
    delete sidt;
    return false;
  }

  sidtRecord *srec=new sidtRecord;
  rec=new tblistRecord;
  if ((srec==NULL) || (rec==NULL)) {
    // no more memory error
    lastError=NO_MEMORY;
    tblistStatus=false;
    if (srec!=NULL) delete srec;
    if (rec!=NULL) delete rec;
    delete sidt;
    return false;
  }

  if (!sidt->getTune(tune, srec)) {
    BDEBUG("Tune " << tune << " is not present\n");
    //tune is not present
    srec->second=0;
    srec->number=0;
    rec->cached=true; 
                                   // use last known path/name          
    rec->path=new char[strlen(lastAbsPath)+1];    
    rec->name=new char[strlen(lastSidName)+1];
    if ((rec->path==NULL) || (rec->name==NULL)) {
      // no more memory error
      lastError=NO_MEMORY;
      tblistStatus=false;
      if (rec->path!=NULL) delete rec->path;
      if (rec->name!=NULL) delete rec->name;
      delete sidt;
      return false;
    }
    strcpy(rec->path, lastAbsPath);
    strcpy(rec->name, lastSidName);
  } else {
       rec->cached=false;
     }

  rec->tune=tune;
  rec->second=srec->second;
  rec->number=srec->number;
  if (isTime) rec->second+=rate;   // update his rate
  else rec->number+=rate;

  delete srec;

  //add MD5
  rec->MD5=new char[strlen(MD5)+1];
  if (rec->MD5==NULL) {
    //no memory error
    lastError=NO_MEMORY;
    tblistStatus=false;
    delete sidt;
    return false;
  }
  strcpy(rec->MD5, MD5);

  // add path/name
  if (rec->cached==false) {                // but avoid if the sidt did not exists
    rec->cached=true;
    rec->path=sidt->getReadSidPath();
    rec->name=sidt->getReadSidName();
  }

  //see if this new value need to enter in the list
  unsigned int min=UINT_MAX;

  /// this if we granted that list is packed
  if (list->number != 0) {
    if (list->item[list->number-1]->cached==false)
      readRecord(list->item[list->number-1]);
    if (isTime) min=list->item[list->number-1]->second;
    else min=list->item[list->number-1]->number;
  }

  // if the value is greater than the min, we update the list
  if (isTime) {
    if (min < rec->second) {
      updateList(list, rec, list->size-1, isTime, false);
      delete sidt;
      return true;
    }
  } else {
      if (min < rec->number) {
        updateList(list, rec, list->size-1, isTime, false);
        delete sidt;
        return true;
      }
    }
  //otherwise, insert the element only if there is free space
  if (list->number!=list->size) {
    BDEBUG("Insert in free space at position " << list->number << endl);
    list->item[list->number]=rec;
    list->number++;
  }

  delete sidt;
  return true;
}

/**
 * Update the passed list inserting the passed record at the right
 * place from position 0 to position pos
 *
 * @param list the list to update
 * @param rec the record to insert
 * @param pos the max position until the list may be updated
 * @param isTime true if must look for time
 * @param skip true if the number updating ma be skipped
 */
void TBLIST::updateList(tblistList *list, tblistRecord *rec,
                        unsigned int pos, bool isTime, bool skip) {
  BDEBUG("Insert a record in the list from pos 0 to " << pos << " with " << skip << endl);
  unsigned int i;

  if (!skip) list->number++;            // suppose this is a new elements

  for (i=0; i<=pos; i++) {
    if (list->item[i]==NULL) break;
    if (list->item[i]->cached==false) readRecord(list->item[i]);
    if (isTime) {
      if (list->item[i]->second < rec->second) break;
    } else {
        if (list->item[i]->number < rec->number) break;
      }
  }

  if (!skip && (list->item[list->size-1]!=NULL)) list->number--;    // this elements is going out

  //shift the elements down
  if (i!=pos) {
    for (unsigned int j=pos; j>i; j--) {
      list->item[j]=list->item[j-1];
    }
  }

  list->item[i]=rec;
  BDEBUG("Element inserted in position: " << i << endl);
}

/**
 * Get the size of the hot list of time
 *
 * @return the hot list time size
 */
unsigned int TBLIST::hotlistTSize() {
  BDEBUG("Get hot list size of second\n");
  if (listTime==NULL) return 0;
  else return listTime->number;
}

/**
 * get the size of the hot list of number of listen
 *
 * @return the hot list number size
 */
unsigned int TBLIST::hotlistNSize() {
  BDEBUG("Get hot list size of number\n");
  if (listNumber==NULL) return 0;
  else return listNumber->number;
}

/**
 * Get a record of information of the tune in time hot list at the passed
 * index (starting from 1).
 * the chached value in the record is used only internally
 *
 * @param index the index of the tune in the list
 * @return the record of information or NULL
 */
const tblistRecord * TBLIST::hotlistTInfo(unsigned int index) {
  BDEBUG("Return " << index << " record from the hot list time\n");

  if (listTime==NULL) return NULL;
  if (index>listTime->number) return NULL;
  if (index==0 || index>listTime->number) return NULL;
  --index;
  if (listTime->item[index]->cached==false) {
    readRecord(listTime->item[index]);
  }
  return listTime->item[index];
}


/**
 * Get a record of information of the tune in number hot list at the passed
 * index (starting from 1).
 * the chached value in the record is used only internally
 *
 * @param index the index of the tune in the list
 * @return the record of information or NULL
 */
const tblistRecord * TBLIST::hotlistNInfo(unsigned int index) {
  BDEBUG("Return " << index << " record from the hot list number\n");

  if (listNumber==NULL) return NULL;
  if (index>listNumber->number) return NULL;
  if (index==0 || index>listNumber->number) return NULL;
  --index;
  if (listNumber->item[index]->cached==false) {
    readRecord(listNumber->item[index]);
  }
  return listNumber->item[index];
}

/**
 * Set the last used absolute name of the sid files being added to the list
 * Used before a call to <code>add</code> to be sure that the if the sidt
 * file did not exist yet (due to buffered tecnique), the name is correct
 *
 * @param absSidFileName the absolute file name
 * @return true if the operation is ok
 */
bool TBLIST::setLast(const char* absSidFileName) {
  BDEBUG("Set last to " << absSidFileName << endl);

  // separate name from path
  int length;                                        // length of result string
  int length2;                                       // another length of string
  char *pnt=(char *)&absSidFileName[strlen(absSidFileName)]; // pointer to the end of string

  while (*pnt!='/') {                                // look for slash
    pnt--;                                           // decrement pointer
    if (pnt<=absSidFileName) break;                  // avoid endless loop
  }

  lastAbsPath=new char[(length=(pnt-absSidFileName))+1];
  if (lastAbsPath==NULL) {
      lastError=NO_MEMORY;                           // no memory available
      tblistStatus=false;
      return false;
  }
  strncpy(lastAbsPath, absSidFileName, length);
  lastAbsPath[length]='\0';                           // terminate the string
  BDEBUG("Last path= " << lastAbsPath << endl);

  lastSidName=new char[(length2=(strlen(absSidFileName)-length))+1];
  if (lastSidName==NULL) {
      lastError=NO_MEMORY;                           // no memory available
      tblistStatus=false;
      return false;
  }
  strncpy(lastSidName, &absSidFileName[length+1], length2);
  BDEBUG("Last name= " << lastSidName << endl);

  return true;
}

/**
 * Get the total listen time in second
 *
 * @return the total listen time in second
 */
unsigned int TBLIST::getTotalTime() {
  BDEBUG("Get the total time: " << totalTime << endl);
  return totalTime;
}

/**
 * Get the total number of listen
 *
 * @return the total number of listen
 */
unsigned int TBLIST::getTotalNumber() {
  BDEBUG("Get the total number of listen: " << totalNumber << endl);
  return totalNumber;
}

/**
 * Get the error string of the last error
 *
 * @return the string of the error
 */
const char* TBLIST::getError() {
  return ERROR_LIST[lastError];
}

#endif // _TBList.CPP


