/**
 * @(#)tsidexaminer2.cpp 2002/02/25
 *
 * @author Ice
 * @version 1.02
 *
 * ICE Team free software group
 *
 * This file is part of TSID2.
 * See README for copyright notice.
 *
 * note: some comments are in the style of Javadoc of Java language!
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 */

#ifndef _TSIDEXAMINER2
#define _TSIDEXAMINER2

#include "tsidexaminer2.h"
#include "sidt2.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <sys/stat.h>

#ifdef HAVE_MSWINDOWS
# define snprintf _snprintf
#endif

#ifdef HAVE_UNIX
#  include <unistd.h>
#  include <sys/fcntl.h>
#  include <dirent.h>
#elif __BORLANDC__
#  include <dirent.h>
#else
#  include "win32-dirent.h"
#  define S_ISDIR(st_mode) (st_mode & S_IFDIR)
#  define S_ISREG(st_mode) (st_mode & S_IFREG)
#endif

#define MAX_PATH_LEN  128
#define NLENGTH(dirent) ((int)strlen((dirent)->d_name))
#define strequ(s1, s2) (strcmp((s1),(s2)) == 0)

const char *TSIDExaminer2::ERR_NO_TSIDE_ERROR   = "TSIDExaminer2: No error";
const char *TSIDExaminer2::ERR_NO_MEMORY        = "TSIDExaminer2: Error allocating memory";
const char *TSIDExaminer2::ERR_NO_BASE_ERROR    = "TSIDExaminer2: No base directory was given";
const char *TSIDExaminer2::ERR_DIR_READ_ERROR   = "TSIDExaminer2: Error reading the directory";
const char *TSIDExaminer2::ERR_TOO_LONG_ERROR   = "TSIDExaminer2: Name too long";
const char *TSIDExaminer2::ERR_EXT_ERROR        = "TSIDExaminer2: External error";
const char *TSIDExaminer2::ERR_INI_OPEN_ERROR   = "TSIDExaminer2: Error opening the INI database";
const char *TSIDExaminer2::ERR_INI_DELIMS_ERROR = "TSIDExaminer2: Error setting delims in the database";
const char *TSIDExaminer2::ERR_INI_READ_ERROR   = "TSIDExaminer2: Error reading the ini file";
const char *TSIDExaminer2::ERROR_LIST[]={
  ERR_NO_TSIDE_ERROR,
  ERR_NO_MEMORY,
  ERR_NO_BASE_ERROR,
  ERR_DIR_READ_ERROR,
  ERR_EXT_ERROR,
  ERR_TOO_LONG_ERROR,
  ERR_INI_OPEN_ERROR,
  ERR_INI_DELIMS_ERROR,
  ERR_INI_READ_ERROR
};

/**
 * Construnct the TSIDExaminer
 *
 * @param absBasePath the base path where there are the sidt2 files
 * @param useIndexMode true if we must use external index file
 */
TSIDExaminer2::TSIDExaminer2(const char *absBasePath, bool useIndexMode) {
  SDEBUG2("Construct the TSIDExaminer2 with: " << absBasePath << endl);
  
  versionString="tsidexaminer2 1.02 by Ice Team Free Software Group (C) 2000-2005";
  
  startList=NULL;
  endList=NULL;
  listPointer=NULL;
  totalNumber=0;
  totalTime=0;
  database=NULL;
  absIndexName=NULL;
  mergeStatus=NOTHING;
  tsidexaminerStatus=true;
  
  // test for base present
  if (absBasePath == NULL) {
    lastError=NO_BASE_ERROR;     // no base path for the stored information
    tsidexaminerStatus=false;    // tsidexaminer2 cannot continue
    return;
  }

  // make a phisical copy of the passed argument
  TSIDExaminer2::absBasePath=new char[strlen(absBasePath)+1];

  if (TSIDExaminer2::absBasePath == NULL) {
    lastError=NO_MEMORY;            // no more memory
    tsidexaminerStatus=false;       // tsidexaminer cannot go away
    return;
  }
  strcpy(TSIDExaminer2::absBasePath, absBasePath);
  
  // remember the mode to use
  TSIDExaminer2::useIndexMode=useIndexMode;
  
  // test if using external index file
  if (useIndexMode) {
    // make the index file name
    absIndexName=new char[strlen(absBasePath)+10];
    
    if (absIndexName == NULL) {
      lastError=NO_MEMORY;            // no more memory
      tsidexaminerStatus=false;       // tsidexaminer2 cannot go away
      return;  
    }
   
    sprintf(absIndexName, "%s/index.ini", absBasePath);
    
    database=ini_open(absIndexName,"r", ";");
    if (!database) {
      SDEBUG2("Error opening the INI database for the " << absIndexName << endl);
      lastError=INI_OPEN_ERROR;
      tsidexaminerStatus=false;       // tsidexaminer2 cannot go away  
      return;
    }
    if (ini_listDelims(database, " ") < 0) {
      SDEBUG2("Error setting delimiters in the INI database\n");
      lastError=INI_DELIMS_ERROR;
      tsidexaminerStatus=false;       // tsidexaminer2 cannot go away
      return;
    }  
  }
}

/**
 * Destruct the TSIDExaminer2
 */
TSIDExaminer2::~TSIDExaminer2() {
  SDEBUG2("Destruct the TSIDExaminer2\n");

  if (absBasePath != NULL) delete[] absBasePath;

  if (startList != NULL) {
    tuneItem *pTmp;                          // temp pointer

    while (startList != NULL) {
      pTmp=startList;                        // remember the start

      delete[] pTmp->rec.MD5;
      delete[] pTmp->rec.path;
      delete[] pTmp->rec.name;

      startList=startList->next;             // the new list
      delete pTmp;                           // release memory
    }
  }
    
  if (database != NULL) {
    ini_close(database);
  }  
  if (absIndexName != NULL) delete[] absIndexName;
}

/**
 * Analize the tsid2 files for retriving the information
 *
 * @return true if the operation is ok
 */
bool TSIDExaminer2::analise() {
  dirent *dp;
  struct stat st;
  DIR    *dirp;
  char   nbuf[MAX_PATH_LEN];
  int    len;
  char   name[37];

  totalNumber=0;
  totalTime=0;

  SDEBUG2("Scan directory " << absBasePath << endl);

  lastError=NO_TSIDE_ERROR;                  // no error
  
  mergeStatus=SUBTUNE;            // we have list of subtunes

  dirp=opendir(absBasePath);

  if (dirp==NULL) {
    SDEBUG2(absBasePath << " unreadable\n");
    lastError=DIR_READ_ERROR;                // read error of directory
    return false;
  }

  while ((dp=readdir(dirp)) != NULL) {
    if (strequ(dp->d_name,".") || strequ(dp->d_name,"..")) {
      continue;
    }
    len=strlen(absBasePath);
    if (len+NLENGTH(dp)+1 < MAX_PATH_LEN-1) {
      strcpy(nbuf, absBasePath);
      if (len != 0) {
        nbuf[len++]='/';
      }
      strcpy(nbuf+len, dp->d_name);

      stat(nbuf, &st);

      if (S_ISREG(st.st_mode) && strlen(dp->d_name)==36 ) { //36=length Md5 + ext
        strcpy(name, dp->d_name);
        if (strcmp(&name[33], "dat")==0) {
          name[32]='\0';
          addFileToList(name);                 // add the file
        }
      }
    } else {
        SDEBUG2(absBasePath << "/" << dp->d_name << ": pathname too long\n");
        lastError=TOO_LONG_ERROR;
      }
  }
  closedir(dirp);

  return lastError==NO_TSIDE_ERROR;
}

/**
 * Insert the information of SIDT2 file into the list
 *
 * @param nfile name of file
 * @return true if the operation is ok
 */
bool TSIDExaminer2::addFileToList(char *nfile) {
  SDEBUG2("Add information from file " << nfile << endl);

  // pointer to sidt file manager
  SIDT2 *pSidt=new SIDT2(nfile, "#/#", absBasePath);

  tuneItem *pItem;               // a item for the list

  if (!*pSidt) {
    SDEBUG2("SIDT never initialize\n");
    lastError=EXT_ERROR;
    dinamicError=pSidt->getError();
    return false;
  }

  if (!pSidt->read()) {
    SDEBUG2("Error reading SIDT information\n");
    dinamicError=pSidt->getError();
    lastError=EXT_ERROR;
    delete pSidt;
    return false;
  }

  pSidt->initForGet();                       // init for getting next subtune

  pItem=new tuneItem;

  // scan all the subtunes that are present
  while (pSidt->getNext(&pItem->rec.tune, &pItem->rec.second,
                        &pItem->rec.number, &pItem->rec.length)) {

    // update total information
    totalNumber+=pItem->rec.number;
    totalTime+=pItem->rec.second;

    // get pondered score
    //getScore(pItem, pSidName);

    SDEBUG2("Add song #" << pItem->rec.tune << endl);

    if (startList==NULL) {                   // list empty ?
      startList=pItem;                       // this is the first element
      endList=pItem;                         // and the last too
    } else {                                 // list not empty
        endList->next=pItem;                 // next point to the last
        endList=pItem;                       // this is the last
      }
    // pItem must be completed with the path and file name
    pItem->rec.MD5=pSidt->getMD5();
    if (!useIndexMode) {
      pItem->rec.path=pSidt->getReadSidPath();
      pItem->rec.name=pSidt->getReadSidName();
    } else {
        int len;
        
        ini_locateHeading(database, "TUNES");
        if (ini_locateKey(database, pItem->rec.MD5)==0) { 
    
          // now read path
          len = ini_dataLength(database) + 1;       
        
          pItem->rec.path=new char[len];
          if (pItem==NULL) {
            delete[] pItem->rec.MD5;
            delete pItem;
        
            lastError=NO_MEMORY;
            tsidexaminerStatus=false;
            return false;             
          }
        
          if (ini_readString(database, pItem->rec.path, len)<0) {
            SDEBUG2("Error reading the path of sid\n");
            lastError=INI_READ_ERROR;
            // try to continue with the other reading
          }        
        
          // now read the name
          len = ini_dataLength(database) + 1;       
        
          pItem->rec.name=new char[len];
          if (pItem==NULL) {
            delete[] pItem->rec.path;
            delete[] pItem->rec.MD5;
            delete pItem;
        
            lastError=NO_MEMORY;
            tsidexaminerStatus=false;
            return false;             
          }       
        
           if (ini_readString(database, pItem->rec.name, len)<0) {
            SDEBUG2("Error reading the name of sid\n");
            lastError=INI_READ_ERROR;
            // try to continue with the other reading
          }
        } else {  // there is not an entry in the database for this tune  
                  // so use the name from the tune
            pItem->rec.path=pSidt->getReadSidPath();
            pItem->rec.name=pSidt->getReadSidName();        
          }            
      }

    pItem->next=NULL;                        // there's not other
    pItem=new tuneItem;
  }

  delete pItem;
  delete pSidt;

  return true;
}

/**
 * Sort the list for the passed type
 *
 * @param type the type of sort to apply
 */
void TSIDExaminer2::sort(sortType type) {
  tuneItem *actual;        // actual pointer
  tuneItem *start;         // actual start pointer (upper, all are ordered)
  tuneItem tmp;            // tmp for swap

  SDEBUG2("Sort List\n");

  // swap element macro
  #define swap(a,b) \
     tmp.rec.path=a->rec.path; \
     tmp.rec.name=a->rec.name; \
     tmp.rec.MD5=a->rec.MD5; \
     tmp.rec.tune=a->rec.tune; \
     tmp.rec.second=a->rec.second; \
     tmp.rec.number=a->rec.number; \
     tmp.rec.length=a->rec.length; \
     a->rec.path=b->rec.path; \
     a->rec.name=b->rec.name; \
     a->rec.MD5=b->rec.MD5; \
     a->rec.tune=b->rec.tune; \
     a->rec.second=b->rec.second; \
     a->rec.number=b->rec.number; \
     a->rec.length=b->rec.length; \
     b->rec.path=tmp.rec.path; \
     b->rec.name=tmp.rec.name; \
     b->rec.MD5=tmp.rec.MD5; \
     b->rec.tune=tmp.rec.tune; \
     b->rec.second=tmp.rec.second; \
     b->rec.number=tmp.rec.number; \
     b->rec.length=tmp.rec.length;

  // use a bubble sort algorithm (oh, what I see)
  if (startList==NULL) return;               // list has no element
  if (startList->next==NULL) return;         // list has one element

  for (start=startList; start->next!=NULL; start=start->next)
    for (actual=start->next; actual!=NULL; actual=actual->next) {
      switch (type) {
        case TIME:
          if (actual->rec.second > start->rec.second) {
             swap(actual, start);
          }
          break;
        case NUMBER:
          if (actual->rec.number > start->rec.number) {
             swap(actual, start);
          }
          break;
       // case SORT_FOR_SCORE:
       //   if (actual->score > start->score) {
       //     swap(actual, start);
       //   }
        // default: do nothing
      }
    }
}

/**
 * Merge the equal song in the list
 * The tunes of the same song/author are merge in one
 */
void TSIDExaminer2::mergeSongList() {
  tuneItem *actual;        // actual pointer
  tuneItem *start;         // actual start pointer (upper, all are merged)
  tuneItem *prev;          // previous element of actual

  SDEBUG2("Merge Song List\n");
  
  // do nothing if we haven't a list
  if (mergeStatus!=SUBTUNE) return;

  // use a bubble sort algorithm (oh, what I see)
  if (startList==NULL) return;               // list has no element
  if (startList->next==NULL) return;         // list has one element

  for (start=startList; start->next!=NULL; start=start->next) {
    for (actual=start->next, prev=start;
         actual!=NULL;
         prev=actual, actual=actual->next) {
      if (strequ(actual->rec.MD5, start->rec.MD5)) {
        // merge the actual element to the start element
        start->rec.second+=actual->rec.second;
        start->rec.number+=actual->rec.number;
        start->rec.length+=actual->rec.length;
        //start->rec.score+=actual->rec.score;
        start->rec.tune=0;

        // remove actual from list
        prev->next=actual->next;
        delete[] actual->rec.path;
        delete[] actual->rec.name;
        delete actual;
        actual=prev;
      }
    }
    // the list is modified, so this is needed
    if (start->next==NULL) break;
  }
  mergeStatus=TUNE;  // we have tune file list
}

/**
 * Merge the songs of the same author in the list
 * The tunes of the same author are merge in one
 * Note: you must call <code>mergeSongList</code> before use this methods.
 */
void TSIDExaminer2::mergeAuthorList() {
  tuneItem *actual;        // actual pointer
  tuneItem *start;         // actual start pointer (upper, all are merged)
  tuneItem *prev;          // previous element of actual
  char *tmpStr="Various";
  char *tmpMD5="00000000000000000000000000000000";

  #define substName \
    delete [] start->rec.name; \
    delete [] start->rec.MD5; \
    start->rec.name=new char[strlen(tmpStr)+1]; \
    start->rec.MD5=new char[strlen(tmpMD5)+1]; \
    if ((start->rec.name==NULL) || (start->rec.MD5==NULL)) { \
      lastError=NO_MEMORY; \
      tsidexaminerStatus=false; \
      return; \
    } \
    strcpy(start->rec.name, tmpStr); \
    strcpy(start->rec.MD5, tmpMD5); \
    start->rec.tune=0;

  SDEBUG2("Merge Author List\n");
  
  // do nothing if we haven't a list with only tune 
  if (mergeStatus!=TUNE) return;

  // use a bubble sort algorithm (oh, what I see)
  if (startList==NULL) return;               // list has no element
  if (startList->next==NULL) {               // list has one element
    start=startList;
    substName;
    return;
  }

  for (start=startList; start->next!=NULL; start=start->next) {
    for (actual=start->next, prev=start;
         actual!=NULL;
         prev=actual, actual=actual->next) {
      if (strequ(actual->rec.name, start->rec.name)) {
        // merge the actual element to the start element
        start->rec.second+=actual->rec.second;
        start->rec.number+=actual->rec.number;
        //start->score+=actual->score;
        start->rec.length+=actual->rec.length;
        // remove actual from list
        prev->next=actual->next;
        delete[] actual->rec.path;
        delete[] actual->rec.name;
        delete[] actual->rec.MD5;
        delete actual;
        actual=prev;
      }
    }
    substName;

    // the list is modified, so this is needed
    if (start->next==NULL) break;
  }
  substName;

  mergeStatus=AUTHOR;   // we have author directory list
}

/**
 * Init the engin for getting the next song information from list
 */
void TSIDExaminer2::initForGet() {
  SDEBUG2("Init engine for getting songs\n");
  
  listPointer=startList;   // start of element in list
}

/**
 * Get next tune information from the list
 * The returned strings are a just created copy
 *
 * @param rec a record of information
 * @return true if the record is present
 */
bool TSIDExaminer2::getNext(struct TRec *rec) {
  SDEBUG2("Get next is called" << endl);
  // finish element or no list available?

  if (listPointer==NULL) {
    return false;
  }

  size_t size=strlen(listPointer->rec.MD5);
  size_t size1=strlen(listPointer->rec.path);
  size_t size2=strlen(listPointer->rec.name);
  rec->MD5=new char[size+1];
  rec->path=new char[size1+1];
  rec->name=new char[size2+1];

  if ((rec->MD5==NULL) || (rec->name==NULL) || (rec->path==NULL)) {
      lastError=NO_MEMORY;
      tsidexaminerStatus=false;
      return false;         // this error stop list creation
  }

  // copy the strings
  strncpy(rec->MD5, listPointer->rec.MD5, size+1);
  strncpy(rec->path, listPointer->rec.path, size1+1);
  strncpy(rec->name, listPointer->rec.name, size2+1);

  SDEBUG2("Get next: " << rec->MD5 << "..." << endl);

  rec->tune=listPointer->rec.tune;
  rec->second=listPointer->rec.second;
  rec->number=listPointer->rec.number;
  rec->length=listPointer->rec.length;

  SDEBUG2("...#" << rec->tune << ", " << rec->second << ", " << rec->number << ", "
          << rec->length << endl);

  listPointer=listPointer->next;

  return true;
}

/**
 * Get the total listen time in second
 *
 * @return the total listen time in second
 */
unsigned int TSIDExaminer2::getTotalTime() {
  SDEBUG2("Get the total time: " << totalTime << endl);
  return totalTime;
}

/**
 * Get the total number of listen
 *
 * @return the total number of listen
 */
unsigned int TSIDExaminer2::getTotalNumber() {
  SDEBUG2("Get the total number of listen: " << totalNumber << endl);
  return totalNumber;
}

/**
 * Get the error string of the last error
 *
 * @return the string of the error
 */
const char* TSIDExaminer2::getError() {
  if (lastError==EXT_ERROR) return dinamicError;
  else return ERROR_LIST[lastError];
}

/**
 * Show the list to the passed output stream
 * The stream must be ready for use
 *
 * @param out the output stream pointer
 * @param withScore true if the score must be displayed
 */
void TSIDExaminer2::showList(ostream *out, bool withScore) {
  tuneItem *actual;              // pointer to the list
  char *tmpStr="Various";
  //int step=strlen(homeTDir);     // this is not to show
  int position=1;                // position in list
  unsigned long totalSec=0;      // total listen seconds
  unsigned long totalNum=0;      // total number of listen

  SDEBUG2("Show list to passed stream\n");

  actual=startList;
  while (actual!=NULL) {
    // calculate statistic
    totalSec+=actual->rec.second;
    totalNum+=actual->rec.number;

    // show the information to the output stream

    *out << position++ << "\t";
    *out << actual->rec.second << "\t";
    *out << actual->rec.number << "\t";

    //if (withScore) *out << (long int)actual->rec.score << "\t";

    //*out << ((actual->pathName)+step) << " ";
    *out << (actual->rec.path) << " ";

    if (strcmp(actual->rec.name, tmpStr)!=0) {
      *out << actual->rec.name << " ";
    }

    if (actual->rec.tune!=0) {
      *out << "#" << actual->rec.tune;
    }

    *out << endl;

    actual=actual->next;                     // go to next element
  }
  *out << "*** STATISTICS ***\n";
  *out << "LISTEN SECONDS: " << totalSec << endl;
  *out << "NUMBER of LISTEN: " << totalNum << endl;
}

#endif // _TSIDExaminer2.CPP


