/**
 * @(#)tblist.h 2002/02/25
 *
 * @author Ice
 * @version 1.02
 *
 * ICE Team free software group
 *
 * This file is part of TSID2.
 * See README for copyright notice.
 *
 * note: some comments are in the style of Javadoc of Java language!
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 */

#ifndef _TBLIST_H
#define _TBLIST_H

#include <libini.h>

#include "sidt2.h"
#include "struct.h"

// show debug string if DEBUG is defined
#ifdef DEBUG_TBLIST
#  define BDEBUG(a) \
    cerr << "[TBLIST] " << a;
#else
#  define BDEBUG(a) ;
#endif

class TSID2;
class TSIDExaminer2;


/**
 * TBLIST class
 *
 * This class manage the real-time generating best listen tunes file.
 * The best dat file contains:
 * <pre>
 *  [HOTLIST_N]
 *  tunes = n
 *  1= MD5 subtune
 *  ...
 *  n= MD5 subtune
 *
 *  [HOTLIST_T]
 *  tunes = m
 *  1= MD5 subtune
 *  ...
 *  m= Md5 subtune
 *
 *  [TOTAL]
 *  time=xxxx
 *  number=yyyy
 * </pre>
 *
 * @author ice
 * @version 1.02 19/12/2004
 */
class TBLIST {
  public:
  
    // Enum that describes the possible errors this class may encounter.
    enum TBLISTerror {
      NO_TBLIST_ERROR = 0,  // no error
      NO_MEMORY,            // error: no more memory
      INI_OPEN_ERROR,       // error opening the INI database
      INI_DELIMS_ERROR,     // error setting delims in the database
      INI_READ_ERROR,       // error reading the INI file
      INI_WRITE_ERROR,      // error writing the INI file
      NO_EXAM_ERROR,        // error creating examiner
      ANALISE_ERROR,        // error analising with examiner
      NO_SIDT_ERROR,        // error creating Sidt
      READ_SIDT_ERROR       // error reading from SIDT
    };

    // enum that describe the type of engine TBLIST can use
    enum engineType {
      STATIC,            // static function: read all the lists at initialization
                         // good for quering the list or update it if the user
                         // listen more tunes in the same section
      DINAMIC            // dinamic function: read the list dinamically as needed
                         // good if user listen one tune for each section
    };


    // the list of pointers to the items
    // the list is packed -> first number items are not empty
    struct tblistList {
      unsigned int number;           // effective number of items in the list
      unsigned int size;             // size of the list
      tblistRecord *item[];          // item of the record
    };

    /**
     * Construct the TBLIST using the passed engine type
     * The path must be with ansi path separator
     *
     * @param engine the type of the engine
     * @param absBasePath the base path where there are the sidt2 files
     * @param number_t the number of time item to store in the list
     * @param number_n the number of number item  to store in the list
     * @param tsid2 a pointer to tsid2 that is using this class
     */
    TBLIST(engineType engine, const char *absBasePath,
           unsigned int number_t, unsigned int number_n, TSID2 *tsid2);

    /**
     * Destruct the TBLIST
     */
    ~TBLIST();
    
    /**
     * Return the state of TBLST: false means that TBLIST cannot continue
     */
    operator bool() {
      return tblistStatus;
    }
    
    /**
     * Get the code of the last error
     */
    inline TBLISTerror getCodeError() {
      return lastError;
    }
    
    /**
     * Get the error string of the last error
     *
     * @return the string of the error
     */
    const char *getError();
    
    /**
     * Get the version of SIDT as a formatted string
     *
     * @return the version of SIDT
     */
    inline char *getVersion() {
      return versionString;
    };

    /**
     * Set the size of the lists
     *
     * @param sizeTime the size that the list with time must take
     * @param sizeNumber the size that the list with number must take
     * @return true if the operation is ok
     */
    bool setSize(unsigned int sizeTime, unsigned int sizeNumber);

    /**
     * Add the passed listen time to the top lists
     *
     * @param MD5 the MD5 of the file tune
     * @param tune the subtune
     * @param second the listen second to add
     * @param number the number of listen time to add
     */
    void add(const char *MD5, unsigned int tune, unsigned long second, unsigned long number=1);

    /**
     * Add the passed listen time to the top lists
     *
     * @param MD5 the MD5 of the file tune
     * @param rec a record of information
     */
    void add(const char *MD5, sidtRecord rec);

    /**
     * Get the size of the hot list of time
     *
     * @return the hot list time size
     */
    unsigned int hotlistTSize();

    /**
     * get the size of the hot list of number of listen
     *
     * @return the hot list number size
     */
    unsigned int hotlistNSize();

    /**
     * Get a record of information of the tune in time hot list at the passed
     * index (starting from 1).
     * the chached value in the record is used only internally
     *
     * @param index the index of the tune in the list
     * @return the record of information or NULL
     */
     const tblistRecord * hotlistTInfo(unsigned int index);

    /**
     * Get a record of information of the tune in number hot list at the passed
     * index (starting from 1).
     * the chached value in the record is used only internally
     *
     * @param index the index of the tune in the list
     * @return the record of information or NULL
     */
     const tblistRecord * hotlistNInfo(unsigned int index);

     /**
      * Set the last used absolute name of the sid files being added to the list
      * Used before a call to <code>add</code> to be sure that the if the sidt
      * file did not exist yet (due to buffered tecnique), the name is correct
      *
      * @param absSidFileName the absolute file name
      * @return true if the operation is ok
      */
     bool setLast(const char* absSidFileName);
     
     /**
      * Get the total listen time in second
      *
      * @return the total listen time in second
      */
     unsigned int getTotalTime();

     /**
      * Get the total number of listen
      *
      * @return the total number of listen
      */
     unsigned int getTotalNumber();
     
     /**
      * Rebuild the two lists by scanning all the stored files
      *
      * @return true if the operation is ok
      */
     bool rebuildLists();

  private:
    static const char *ERR_NO_TBLIST_ERROR;
    static const char *ERR_NO_MEMORY;
    static const char *ERR_INI_OPEN_ERROR;
    static const char *ERR_INI_DELIMS_ERROR;
    static const char *ERR_INI_READ_ERROR;
    static const char *ERR_INI_WRITE_ERROR;
    static const char *ERR_NO_EXAM_ERROR;
    static const char *ERR_ANALISE_ERROR;
    static const char *ERR_NO_SIDT_ERROR;
    static const char *ERR_READ_SIDT_ERROR;

    static const char *ERROR_LIST[];

    /**
     * Create an empty list with the passed size
     *
     * @param size the size of the list
     * @return the pointer to the list
     */
    tblistList *createEmptyList(unsigned int size);

    /**
     * Delete the passed list by free the used memory
     *
     * @param list the pointer to the list
     */
    void deleteList(tblistList *list);

    /**
     * Read the lists info from the file
     *
     * @return true if the operation is ok
     */
    bool readLists();

    /**
     * Read the item list from the file
     *
     * @param heading the heading string
     * @param list the list to complete with the entries
     * @param maxSize the max size allocated for this object
     * @return true if the operation is ok
     */
    bool readList(char *heading, tblistList *list, unsigned int maxSize);

    /**
     * Read the record of the list from the file
     * The MD5 and tune field of the record must be filled with the data to read
     * before calling this routine
     *
     * @param rec the record to read in
     * @return true if the operation is ok
     */
    bool readRecord(tblistRecord *rec);

    /**
     * Reduce the size of the passed list by creating a new short list.
     * The passed list will be erased
     *
     * @param list the list to reduce
     * @param size the new shorter size
     * @return a new reduced list
     */
    tblistList *reduceList(tblistList *list, unsigned int size);

    /**
     * Write the lists info to the file
     *
     * @return true if the operation is ok
     */
    bool writeLists();

    /**
     * Write the value of the passed list in the database file
     *
     * @param list the pointer to the list
     * @param head the heading to write in the database
     * @return true if the operation is ok
     */
    bool writeList(tblistList *list, char *head);

    /**
     * Add the passed rate to the given list at the right position
     *
     * @param list the list to update
     * @param MD5 the MD5 of the file
     * @param tune the subtune to update
     * @param rate the rate to add to the current value
     * @param isTime true if this is the time to update
     * @return true if the operation is ok
     */
    bool addList(tblistList *list, const char *MD5, unsigned int tune,
                 unsigned long int rate, bool isTime);

    /**
     * Update the passed list inserting the passed record at the right
     * place from position 0 to position pos
     *
     * @param list the list to update
     * @param rec the record to insert
     * @param pos the max position until the list may be updated
     * @param isTime true if must look for time
     * @param skip true if the number updating ma be skipped
     */
    void updateList(tblistList *list, tblistRecord *rec, unsigned int pos, bool isTime, bool skip);
    
    /**
     * Rebuild the time list
     *
     * @param sizeTime the size that the list with time must take
     * @param exam an already initialized TSIDExaminer2 object
     */
    void rebuildListT(unsigned int sizeTime, TSIDExaminer2 *exam);

    /**
     * Rebuild the number list
     *
     * @param sizeNumber the size that the list with number must take
     * @param exam an already initialized TSIDExaminer2 object
     */
    void rebuildListN(unsigned int sizeNumber, TSIDExaminer2 *exam);
    
    // Error number of the last error that happened.
    TBLISTerror lastError;
    
    // true if tblist can continue to work correctly
    bool tblistStatus;

    // Version number/copyright string
    char *versionString;

    // The type of engine to use
    engineType engine;

    // The absolute base path where there are the sidt2 files
    char *absBasePath;

    // The absolute INI file name with the lists
    char *absFileName;

    // the last known abs path of the sid file
    char *lastAbsPath;

    // the last known abs name of the sid file
    char *lastSidName;

    // The number of time item to store in the list
    unsigned int number_t;

    // The number of number item to store in the list
    unsigned int number_n;

    // The list of best time listen tunes
    tblistList *listTime;

    // The list of best number of listen tunes
    tblistList *listNumber;

    // The total number of listen
    unsigned int totalNumber;

    // The total listen time
    unsigned int totalTime;

    // INI datadase of the tblist file
    ini_fd_t database;

    // the pointer to tsid2 that is using this class
    TSID2 *tsid2;

};

#endif // _TBLIST_H
